/*
 * ft2tf - truetype to tinyfont convertor
 *
 * Copyright (C) 2009-2014 Ali Gholami Rudi <ali at rudi dot ir>
 *
 * This program is released under the Modified BSD license.
 */
#include <unistd.h>
#include <string.h>

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_BITMAP_H
#include "chars.h"

#define NFONTS		16
#define DWCHAR		0x40000000u	/* 2nd half of a fullwidth char */

static int hdpi = 196, vdpi = 196;	/* horizontal and vertical DPI */
static int autohint;
static int embolden;

#define MAXDOTS		1024
#define MAXGLYPHS	(1 << 16)

#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) > (b) ? (a) : (b))
#define LIMIT(n, a, b)	((n) < (a) ? (a) : ((n) > (b) ? (b) : (n)))

static FT_Library library;
static FT_Face faces[NFONTS];
static int desc[NFONTS];
static int nfaces;
static int rows, cols;

static void xdie(char *msg)
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

static int glyphload(FT_Face face, int c)
{
	int flags = FT_LOAD_RENDER;
	if (autohint)
		flags |= FT_LOAD_FORCE_AUTOHINT;
	return !FT_Get_Char_Index(face, c) ||
			FT_Load_Char(face, c, flags);
}

static int facedescender(FT_Face face)
{
	char *s = "gy_pj/\\|Q";
	int ret = 0;
	for (; *s; s++)
		if (!glyphload(face, *s))
			ret = MAX(ret, face->glyph->bitmap.rows -
					face->glyph->bitmap_top);
	return ret;
}

static void font_add(char *path, float size)
{
	FT_Face *face = &faces[nfaces];
	if (FT_New_Face(library, path, 0, face))
		xdie("failed to load font");
	if (FT_Set_Char_Size(*face, 0, size * 64, hdpi, vdpi))
		xdie("cannot set font size");
	desc[nfaces] = facedescender(*face);
	nfaces++;
}

static unsigned char *font_bitmap(int c)
{
	static unsigned char bits[MAXDOTS];
	int sr, sc;	/* the starting and the number of rows in bits */
	int nr, nc;	/* the starting and the number of cols in bits */
	int br;		/* the starting column of glyph bitmap */
	int bc;		/* the starting row of glyph bitmap */
	int bw;		/* the number of columns in glyph bitmap */
	int i, j;
	unsigned char *src;
	FT_Face face = NULL;
	for (i = 0; i < nfaces; i++) {
		if (!glyphload(faces[i], c & ~DWCHAR)) {
			face = faces[i];
			break;
		}
	}
	if (!face)
		return NULL;
	bc = c & DWCHAR ? cols : 0;
	nc = LIMIT(face->glyph->bitmap.width - bc, 0, cols);
	sc = LIMIT(face->glyph->bitmap_left - bc, 0, cols - nc);
	sr = rows - desc[i] - face->glyph->bitmap_top;
	br = MAX(0, -sr);
	sr = LIMIT(sr, 0, rows);
	nr = MIN(rows - sr, face->glyph->bitmap.rows - br);
	memset(bits, 0, rows * cols);
	src = face->glyph->bitmap.buffer;
	bw = face->glyph->bitmap.pitch;
	if (face->glyph->bitmap.pixel_mode == FT_PIXEL_MODE_MONO) {
		for (i = 0; i < nr; i++) {
			for (j = 0; j < nc; j++) {
				int num = (br + i) * bw * 8 + bc + j;
				int val = src[num / 8] & (1 << (7 - num % 8));
				bits[(i + sr) * cols + j + sc] = val ? 255 : 0;
			}
		}
		return bits;
	}
	for (i = 0; i < nr; i++)
		memcpy(&bits[(sr + i) * cols + sc], src + (br + i) * bw + bc, nc);
	return bits;
}

static void font_free(void)
{
	int i;
	for (i = 0; i < nfaces; i++)
		FT_Done_Face(faces[i]);
	FT_Done_FreeType(library);
}

struct tf_header {
	char sig[8];
	int ver;
	int n;
	int rows, cols;
};

int isdw(int c);

static int get_glyphs(int *glyphs)
{
	int i, j;
	int n = 0;
	for (i = 0; i < sizeof(chars) / sizeof(chars[0]); i++) {
		for (j = chars[i][0]; j <= chars[i][1] && n < MAXGLYPHS; j++) {
			if (font_bitmap(j))
				glyphs[n++] = j;
			if (font_bitmap(j) && isdw(j))
				glyphs[n++] = DWCHAR | j;
		}
	}
	return n;
}

static unsigned char *bitmap(int c)
{
	unsigned char *bits = font_bitmap(c);
	int i, j;
	if (embolden) {
		for (i = 0; i < rows; i++) {
			for (j = cols - 2; j >= 0; j--) {
				int idx = i * cols + j;
				bits[idx + 1] = MAX(bits[idx], bits[idx + 1]);
			}
		}
	}
	return bits;
}

static int intcmp(void *v1, void *v2)
{
	return *(int *) v1 - *(int *) v2;
}

/*
 * tiny font format:
 * - each file starts with tinyfont[version]
 * - [number of glyphs][rows][cols]
 * - a list of [the ids of a glyph]
 * - a list of [a stream of size rows * cols describing a glyph]
 */
static void tf_write(int fd)
{
	char *sig = "tinyfont";
	struct tf_header head;
	int glyphs[MAXGLYPHS];
	int i;
	memcpy(head.sig, sig, strlen(sig));
	head.ver = 0;
	head.rows = rows;
	head.cols = cols;
	head.n = get_glyphs(glyphs);
	qsort(glyphs, head.n, sizeof(glyphs[0]), (void *) intcmp);
	write(fd, &head, sizeof(head));
	write(fd, glyphs, sizeof(*glyphs) * head.n);
	for (i = 0; i < head.n; i++)
		write(fd, bitmap(glyphs[i]), rows * cols);
	font_free();
	fprintf(stderr, "tinyfont[%5d]: height=%2d width=%2d\n",
		head.n, rows, cols);
}

static void setdpi(char *s)
{
	hdpi = atoi(s);
	vdpi = strchr(s, ':') ? atoi(strchr(s, ':') + 1) : hdpi;
}

static char *usage =
	"usage: ft2tf [options] font1.ttf:size ... >font.tf\n"
	"\noptions:\n"
	"  -h n  \t\t set glyph height\n"
	"  -w n  \t\t set glyph width\n"
	"  -a    \t\t force autohint option\n"
	"  -b    \t\t embolden\n"
	"  -d v:h\t\t set vertical and horizontal DPI\n"
	"  -v    \t\t show version\n";

int main(int argc, char *argv[])
{
	int i;
	char *wdiff = NULL;
	char *hdiff = NULL;
	for (i = 1; i < argc; i++) {
		if (argv[i][0] != '-')
			break;
		else if (argv[i][1] == 'v')
			printf("ft2tf-0.9\n");
		else if (argv[i][1] == 'a')
			autohint = 1;
		else if (argv[i][1] == 'b')
			embolden = 1;
		else if (argv[i][1] == 'd')
			setdpi(argv[i][2] ? argv[i] + 2 : argv[++i]);
		else if (argv[i][1] == 'w')
			wdiff = argv[i][2] ? argv[i] + 2 : argv[++i];
		else if (argv[i][1] == 'h')
			hdiff = argv[i][2] ? argv[i] + 2 : argv[++i];
		else
			fprintf(stderr, usage);
	}
	if (i >= argc) {
		printf(usage);
		return 0;
	}
	FT_Init_FreeType(&library);
	for (; i < argc; i++) {
		char *name = argv[i];
		float size = 10;
		if (strchr(name, ':')) {
			size = atof(strrchr(name, ':') + 1);
			strrchr(name, ':')[0] = '\0';
		}
		font_add(name, size);
	}
	rows = faces[0]->size->metrics.height >> 6;
	cols = faces[0]->size->metrics.max_advance >> 6;
	if (hdiff)
		rows = strchr("-+", hdiff[0]) ? rows + atoi(hdiff) : atoi(hdiff);
	if (wdiff)
		cols = strchr("-+", wdiff[0]) ? cols + atoi(wdiff) : atoi(wdiff);
	tf_write(1);
	return 0;
}
